class Section(object):
    def __init__(self, doc, category, title_start, title_end, body_start, body_end, parent=None, rule=None):
        self.doc = doc
        self.category = category
        self.title_start = title_start
        self.title_end = title_end
        self.body_start = body_start
        self.body_end = body_end
        self.parent = parent
        self.rule = rule

    def __repr__(self):
        return f"""Section(category={self.category}, title={self.title_span}, body={self.body_span}, parent={self.parent}, rule={self.rule})"""

    @property
    def title_span(self):
        return self.doc[self.title_start : self.title_end]

    @property
    def body_span(self):
        return self.doc[self.body_start : self.body_end]

    @property
    def section_span(self):
        return self.doc[self.title_start : self.body_end]



from sys import platform

from medspacy.visualization import visualize_ent
import spacy

DEFAULT_PIPENAMES = {
    "sentencizer",
    "target_matcher",
    "context",
    "tokenizer",
}

ALL_PIPE_NAMES = {
    "sentencizer",
    "target_matcher",
    "context",
    "tokenizer",
    "preprocessor",
    "postprocessor",
    "sectionizer"
}


def load(model="default", enable=None, disable=None, load_rules=True, quickumls_path = None):
    """Load a spaCy language object with medSpaCy pipeline components.
    By default, the base model will be a blank 'en' model with the
    following components:
        - "tokenizer": A customized tokenizer (set to be nlp.tokenizer)
        - "sentencizer": PyRuSH Sentencizer for sentence splitting
        - "target_matcher": TargetMatcher for extended pattern matching
        - "context": ConText for attribute assertion
    Args:
        model: (str or spaCy Lang model) The base spaCy model to load.
            If 'default', will instantiate from a blank 'en' model.
            Otherwise, if it is a string it will call `spacy.load(model)` along with the enable and disable
                arguments and then add medspaCy pipeline components.
            If it is a spaCy language model, then it will simply add medspaCy components to the existing pipeline.
            Examples:
                >>> nlp = medspacy.load()
                >>> nlp = medspacy.load("en_core_web_sm", disable={"ner"})
                >>> nlp = spacy.load("en_core_web_sm", disable={"ner"}); nlp = medspacy.load(nlp)
        enable (iterable, str, or None): A string or list of component names to include in the pipeline.
            If None, will include all pipeline components listed above.
            If "all", will load all medspaCy components.
            If a list or other iterable, will load the specified pipeline components.
            Note that if using enable, *all* desired pipeline component names must be included.
            Pipeline components could also be instantiated separately and added using the `nlp.add_pipe` method.
            In addition to the default values above, the following medspaCy components may also be added
            and will be added if enable="all":
                - "sectionizer": A SectionDetection component.
                    See medspacy.section_detection.Sectionizer
                - "preprocessor": A wrapper around the tokenizer for destructive preprocessing.
                    Rules added to the preprocessor will modify the underlying text.
                    This component will be set as nlp.tokenizer and will not be listed with nlp.pipe_names.
                    See medspacy.preprocess.Preprocessor
                - "postprocessor": A component for implementing custom business logic at the end of the pipeline
                    and modifying entities by removing them from doc.ents or setting attributes.
                    See medspacy.postprocess.Postprocessor
            Any additional component names (ie., not a medspaCy component) will be passed into
                spacy.load(model_name, enable=enable) and will apply to the base model.
                ie., medspacy.load("en_core_web_sm", enable=["tagger", "parser", "context"])
                will load the tagger and parser from en_core_web_sm and then add context.
        disable (iterable or None): A list of component names to exclude.
            Cannot be set if `enable` is not None.
        load_rules (bool): Whether or not to include default rules for available components.
            If True, sectionizer and context will both be loaded with default rules.
            Default is True.
        quickumls_path (string or None): Path to QuickUMLS resource

    Returns:
        nlp: a spaCy Language object
    """
    enable, disable = _build_pipe_names(enable, disable)
    import spacy
    if isinstance(model, str):
        if model == "default":
            nlp = spacy.blank("en")
        else:
            nlp = spacy.load(model, disable=disable)
    # Check if it is a spaCy model
    elif "spacy.lang" in str(type(model)):
        nlp = model
    else:
        raise ValueError("model must be either 'default', the string name of a spaCy model, or an actual spaCy model. "
                         "You passed in", type(model))

    if "tokenizer" in enable:
        from .custom_tokenizer import create_medspacy_tokenizer
        medspacy_tokenizer = create_medspacy_tokenizer(nlp)
        nlp.tokenizer = medspacy_tokenizer

    if "preprocessor" in enable:
        from .preprocess import Preprocessor
        preprocessor = Preprocessor(nlp.tokenizer)
        nlp.tokenizer = preprocessor


    if "sentencizer" in enable:
        from os import path
        from pathlib import Path

        pyrush_path = path.join(
            Path(__file__).resolve().parents[1], "resources", "rush_rules.tsv"
        )
        from .sentence_splitting import PyRuSHSentencizer

        pyrush = PyRuSHSentencizer(pyrush_path)
        # If there is a dependency parser already in the pipeline,
        # adding a sentencizer after will throw an error
        if "parser" in nlp.pipe_names:
            if "tagger" in nlp.pipe_names:
                nlp.add_pipe(pyrush, before="tagger")
            else:
                nlp.add_pipe(pyrush, before="parser")
        else:
            nlp.add_pipe(pyrush)

    if "target_matcher" in enable:
        from .ner import TargetMatcher

        target_matcher = TargetMatcher(nlp)
        nlp.add_pipe(target_matcher)
        
    if "quickumls" in enable:
        from os import path
        from pathlib import Path

        if quickumls_path is None:
            # let's use a default sample that we provide in medspacy
            # NOTE: Currently QuickUMLS uses an older fork of simstring where databases
            # cannot be shared between Windows and POSIX systems so we distribute the sample for both:

            quickumls_platform_dir = 'QuickUMLS_SAMPLE_lowercase_POSIX_unqlite'
            if platform.startswith("win"):
                quickumls_platform_dir = 'QuickUMLS_SAMPLE_lowercase_Windows_unqlite'

            quickumls_path = path.join(
                Path(__file__).resolve().parents[1], "resources", "quickumls/{0}".format(quickumls_platform_dir)
            )
            print('Loading QuickUMLS resources from a default SAMPLE of UMLS data from here: {}'.format(quickumls_path))

        from quickumls.spacy_component import SpacyQuickUMLS

        quickumls_component = SpacyQuickUMLS(nlp, quickumls_path)
        nlp.add_pipe(quickumls_component)

    if "context" in enable:
        from .context import ConTextComponent

        if load_rules:
            context = ConTextComponent(nlp, rules="default")
        else:
            context = ConTextComponent(nlp, rules=None)
        nlp.add_pipe(context)

    if "sectionizer" in enable:
        from .section_detection import Sectionizer

        if load_rules:
            sectionizer = Sectionizer(nlp, rules="default")
        else:
            sectionizer = Sectionizer(nlp, rules=None)
        nlp.add_pipe(sectionizer)

    if "postprocessor" in enable:
        from .postprocess import Postprocessor
        postprocessor = Postprocessor()
        nlp.add_pipe(postprocessor)

    return nlp


def _build_pipe_names(enable=None, disable=None):
    """Implement logic based on the pipenames defined in 'enable' and 'disable'.
    If enable and disable are both None, then it will load the default pipenames.
    Otherwise, will allow custom selection of components.
    """
    if enable is not None and disable is not None:
        raise ValueError("Either `enable` or `disable` must be None.")

    if disable is not None:
        # If there's a single pipe name, next it in a set
        if isinstance(disable, str):
            disable = {disable}
        else:
            disable = set(disable)
        enable = DEFAULT_PIPENAMES.difference(set(disable))
    elif enable is not None:

        if isinstance(enable, str):
            if enable == "all":
                enable = ALL_PIPE_NAMES.copy()
            else:
                enable = {enable}
        else:
            enable = set(enable)
        disable = set(DEFAULT_PIPENAMES).difference(enable)
    else:
        enable = DEFAULT_PIPENAMES
        disable = set()

    return enable, disable



class ConTextItem:

    def __init__(
        self,
        *args,
        **kwargs
    ):
        """
        Warning: Deprecated.
        ConTextItem has been replaced with ConTextRule:
            >>> from medspacy.context import ConTextRule
            >>> rule = ConTextRule("no evidence of", "NEGATED_EXISTENCE", direction="FORWARD")

        """
        raise NotImplementedError("ConTextItem has been deprecated and replaced with ConTextRule. "
                                  "Please import ConTextRule as: "
                                  "`from medspacy.context import ConTextRule`;"
                                  " ConTextRule(literal, category, pattern, direction)")


from spacy.tokens import Doc, Token, Span

# Filepath to default rules which are included in package
from os import path
from pathlib import Path
import re
import warnings

from . import util
from .section_rule import SectionRule
from .section import Section
from ..common.medspacy_matcher import MedspacyMatcher

DEFAULT_RULES_FILEPATH = path.join(Path(__file__).resolve().parents[2], "resources", "section_patterns.json",)

DEFAULT_ATTRS = {
    "past_medical_history": {"is_historical": True},
    "sexual_and_social_history": {"is_historical": True},
    "family_history": {"is_family": True},
    "patient_instructions": {"is_hypothetical": True},
    "education": {"is_hypothetical": True},
    "allergy": {"is_hypothetical": True},
}


class Sectionizer:
    name = "sectionizer"

    def __init__(
        self,
        nlp,
        rules="default",
        add_attrs=False,
        max_scope=None,
        include_header=False,
        phrase_matcher_attr="LOWER",
        require_start_line=False,
        require_end_line=False,
        newline_pattern=r"[\n\r]+[\s]*$",
    ):
        """Create a new Sectionizer component. The sectionizer will search for spans in the text which
        match section header rules, such as 'Past Medical History:'. Sections will be represented
        in custom attributes as:
            category (str): A normalized title of the section. Example: 'past_medical_history'
            section_title (Span): The Span of the doc which was matched as a section header.
                Example: 'Past Medical History:'
            section_span (Span): The entire section of the note, starting with section_header and up until the end
                of the section, which will be either the start of the next section header of some pre-specified
                scope. Example: 'Past Medical History: Type II DM'

        Section attributes will be registered for each Doc, Span, and Token in the following attributes:
            Doc._.sections: A list of namedtuples of type Section with 4 elements:
                - section_title
                - section_header
                - section_parent
                - section_span.
            A Doc will also have attributes corresponding to lists of each
                (ie., Doc._.section_titles, Doc._.section_headers, Doc._.section_parents, Doc._.section_list)
            (Span|Token)._.section_title
            (Span|Token)._.section_header
            (Span|Token)._.section_parent
            (Span|Token)._.section_span

        Args:
            nlp: A SpaCy language model object
            rules (str, list, or None): Where to read rules from. Default is "default", which will
                load the default rules provided by medSpaCy, which are derived from MIMIC-II.
                If a list, should be a list of pattern dicts following these conventional spaCy formats:
                    [
                        {"section_title": "past_medical_history", "pattern": "Past Medical History:"},
                        {"section_title": "problem_list", "pattern": [{"TEXT": "PROBLEM"}, {"TEXT": "LIST"}, {"TEXT": ":"}]}
                    ]
                If a string other than "default", should be a path to a jsonl file containing rules.
            max_scope (None or int): Optional argument specifying the maximum number of tokens following a section header
                which can be included in a section. This can be useful if you think your section rules are incomplete
                and want to prevent sections from running too long in the note. Default is None, meaning that the scope
                of a section will be until either the next section header or the end of the document.
            include_title (bool): whether the section title is included in the section text
            phrase_matcher_attr (str): The name of the token attribute which will be used by the PhraseMatcher
                for any rules with a "pattern" value of a string.
            require_start_line (bool): Optionally require a section header to start on a new line. Default False.
            require_end_line (bool): Optionally require a section header to end with a new line. Default False.
            newline_pattern (str): Regular expression to match the new line either preceding or following a header
                if either require_start_line or require_end_line are True.
        """
        self.nlp = nlp
        self.add_attrs = add_attrs
        self.matcher = MedspacyMatcher(nlp, phrase_matcher_attr=phrase_matcher_attr)
        self.max_scope = max_scope
        self.require_start_line = require_start_line
        self.require_end_line = require_end_line
        self.newline_pattern = re.compile(newline_pattern)
        self.assertion_attributes_mapping = None
        self._parent_sections = {}
        self._parent_required = {}
        self._rule_item_mapping = self.matcher._rule_item_mapping
        self._rules = []
        self.include_header = include_header

        if rules is not None:
            if rules == "default":
                import os

                if not os.path.exists(DEFAULT_RULES_FILEPATH):
                    raise FileNotFoundError(
                        "The expected location of the default rules file cannot be found. Please either "
                        "add rules manually or add a jsonl file to the following location: ",
                        DEFAULT_RULES_FILEPATH,
                    )
                self.add(SectionRule.from_json(DEFAULT_RULES_FILEPATH))
            # If a list, add each of the rules in the list
            elif isinstance(rules, list):
                self.add(rules)
            elif isinstance(rules, str):
                path.exists(rules)
                self.add(SectionRule.from_json(rules))

        if add_attrs is False:
            self.add_attrs = False
        elif add_attrs is True:
            self.assertion_attributes_mapping = DEFAULT_ATTRS
            self.register_default_attributes()
        elif isinstance(add_attrs, dict):
            # Check that each of the attributes being added has been set
            for modifier in add_attrs.keys():
                attr_dict = add_attrs[modifier]
                for attr_name, attr_value in attr_dict.items():
                    if not Span.has_extension(attr_name):
                        raise ValueError("Custom extension {0} has not been set. Call Span.set_extension.")

            self.add_attrs = True
            self.assertion_attributes_mapping = add_attrs

        else:
            raise ValueError("add_attrs must be either True (default), False, or a dictionary, not {0}".format(add_attrs))

    @property
    def rules(self):
        """Returns list of ConTextItems"""
        return self._rules


    def register_default_attributes(self):
        """Register the default values for the Span attributes defined in DEFAULT_ATTRS."""
        for attr_name in [
            "is_negated",
            "is_uncertain",
            "is_historical",
            "is_hypothetical",
            "is_family",
        ]:
            try:
                Span.set_extension(attr_name, default=False)
            except ValueError:  # Extension already set
                pass

    def add(self, rules):
        """Add a list of SectionRules to the clinical_sectionizer.
       Example:
       >>> rules = [ \
            SectionRule("pmh", "past_medical_history"),\
            SectionRule("pmh", "past_medical_history", \
                pattern=[{"LOWER": "past","OP": "?"}, {"LOWER":"medical"}, {"LOWER": "history"}]),\
            SectionRule("a/p:", "assessment_and_plan", pattern=r"a[/&]p:")]\
           ]
       >>> sectionizer.add(rules)
       """
        if not isinstance(rules, list):
            rules = [rules]

        if not isinstance(rules[0], SectionRule):
            if isinstance(rules[0], dict):
                raise TypeError("Dictionary patterns are no longer supported. You should now add rules using the "
                                "`SectionRule` class: `from medspacy.section_detection import SectionRule`. "
                                "You can migrate old patterns to the new rule format by using: "
                                "`from medspacy.section_dection import section_patterns_to_rules; "
                                "rules = section_patterns_to_rules(patterns)`")
            else:
                raise TypeError("Rules must be of class SectionRule, not", type(rules[0]))

        self.matcher.add(rules)

        for rule in rules:
            name = rule.category
            parents = rule.parents
            parent_required = rule.parent_required

            if name in self._parent_sections.keys() and parents != []:
                warnings.warn(
                    "Duplicate section title {0}. Merging parents. If this is not indended, please specify distinc titles.".format(
                        name
                    ),
                    RuntimeWarning,
                )
                self._parent_sections[name].update(parents)
            else:
                self._parent_sections[name] = set(parents)

            if name in self._parent_required.keys() and self._parent_required[name] != parent_required:
                warnings.warn(
                    "Duplicate section title {0} has different parent_required option. Setting parent_required to False.".format(
                        name
                    ),
                    RuntimeWarning,
                )
                self._parent_required[name] = False
            else:
                self._parent_required[name] = parent_required

            self._rules.append(rule)

    def set_parent_sections(self, sections):
        """Determine the legal parent-child section relationships from the list
        of in-order sections of a document and the possible parents of each
        section as specified during direction creation.

        Args:
            sections: a list of spacy match tuples found in the doc
        """
        sections_final = []
        removed_sections = 0
        for i, (match_id, start, end) in enumerate(sections):
            name = self._rule_item_mapping[self.nlp.vocab.strings[match_id]].category
            required = self._parent_required[name]
            i_a = i - removed_sections  # adjusted index for removed values
            if required and i_a == 0:
                removed_sections += 1
                continue
            elif i_a == 0 or name not in self._parent_sections.keys():
                sections_final.append((match_id, start, end, None))
            else:
                parents = self._parent_sections[name]
                identified_parent = None
                for parent in parents:
                    # go backwards through the section "tree" until you hit a root or the start of the list
                    candidate = self._rule_item_mapping[self.nlp.vocab.strings[sections_final[i_a - 1][0]]].category
                    candidates_parent_idx = sections_final[i_a - 1][3]
                    if candidates_parent_idx is not None:
                        candidates_parent = self._rule_item_mapping[
                            self.nlp.vocab.strings[sections_final[candidates_parent_idx][0]]
                        ].category
                    else:
                        candidates_parent = None
                    candidate_i = i_a - 1
                    while candidate:
                        if candidate == parent:
                            identified_parent = candidate_i
                            candidate = None
                        else:
                            # if you are at the end of the list... no parent
                            if candidate_i < 1:
                                candidate = None
                                continue
                            # if the current candidate has no parent... no parent exists
                            if not candidates_parent:
                                candidate = None
                                continue
                            # otherwise get the previous item in the list
                            temp = self._rule_item_mapping[self.nlp.vocab.strings[sections_final[candidate_i - 1][0]]].category
                            temp_parent_idx = sections_final[candidate_i - 1][3]
                            if temp_parent_idx is not None:
                                temp_parent = self._rule_item_mapping[
                                    self.nlp.vocab.strings[sections_final[temp_parent_idx][0]]
                                ].category
                            else:
                                temp_parent = None
                            # if the previous item is the parent of the current item
                            # OR if the previous item is a sibling of the current item
                            # continue to search
                            if temp == candidates_parent or temp_parent == candidates_parent:
                                candidate = temp
                                candidates_parent = temp_parent
                                candidate_i -= 1
                            # otherwise, there is no further tree traversal
                            else:
                                candidate = None

                # if a parent is required, then add
                if identified_parent or not required:
                    # if the parent is identified, add section
                    # if the parent is not required, add section
                    # if parent is not identified and required, do not add the section
                    sections_final.append((match_id, start, end, identified_parent))
                else:
                    removed_sections += 1
        return sections_final

    def set_assertion_attributes(self, ents):
        """Add Span-level attributes to entities based on which section they occur in.

        Args:
            edges: the edges to modify

        """
        for ent in ents:
            if ent._.section.category in self.assertion_attributes_mapping:
                attr_dict = self.assertion_attributes_mapping[ent._.section.category]
                for (attr_name, attr_value) in attr_dict.items():
                    setattr(ent._, attr_name, attr_value)

    def __call__(self, doc):
        matches = self.matcher(doc)
        if self.require_start_line:
            matches = self.filter_start_lines(doc, matches)
        if self.require_end_line:
            matches = self.filter_end_lines(doc, matches)
        matches = self.set_parent_sections(matches)

        # If this has already been processed by the sectionizer, reset the sections
        doc._.sections = []
        # if there were no matches, return the doc as one section
        if len(matches) == 0:
            doc._.sections.append(Section(doc, None, 0, 0, 0, len(doc)))
            return doc

        section_list = []
        # if the firt match does not begin at token 0, handle the first section
        first_match = matches[0]
        if first_match[1] != 0:
            section_list.append(Section(doc, None, 0, 0, 0, first_match[1]))

        # handle section spans
        for i, match in enumerate(matches):
            (match_id, start, end, parent_idx) = match
            if parent_idx is not None:
                parent = section_list[parent_idx]
            else:
                parent = None
            rule = self._rule_item_mapping[self.nlp.vocab.strings[match_id]]
            category = self._rule_item_mapping[self.nlp.vocab.strings[match_id]].category
            # If this is the last match, it should include the rest of the doc
            if i == len(matches) - 1:
                if self.max_scope is None:
                    section_list.append(Section(doc, category, start, end, end, len(doc), parent, rule))
                else:
                    scope_end = min(end + self.max_scope, doc[-1].i)
                    section_list.append(Section(doc, category, start, end, end, scope_end, parent, rule))
            # Otherwise, go until the next section header
            else:
                next_match = matches[i + 1]
                _, next_start, _, _ = next_match
                if self.max_scope is None:
                    section_list.append(Section(doc, category, start, end, end, next_start, parent, rule))
                else:
                    scope_end = min(end + self.max_scope, next_start)
                    section_list.append(Section(doc, category, start, end, end, scope_end, parent, rule))

        for section in section_list:
            doc._.sections.append(section)
            for token in section.section_span:
                token._.section = section


        # If it is specified to add assertion attributes,
        # iterate through the entities in doc and add them
        if self.add_attrs is True:
            self.set_assertion_attributes(doc.ents)
        return doc

    def filter_start_lines(self, doc, matches):
        "Filter a list of matches to only contain spans where the start token is the beginning of a new line."
        return [m for m in matches if util.is_start_line(m[1], doc, self.newline_pattern)]

    def filter_end_lines(self, doc, matches):
        "Filter a list of matches to only contain spans where the start token is followed by a new line."
        return [m for m in matches if util.is_end_line(m[2] - 1, doc, self.newline_pattern)]



"""Module for utility functions contained by multiple components of medspaCy."""
import re

def span_contains(span, target, regex=True, case_insensitive=True):
    """Return True if a Span object contains a target phrase.
    Case insensitive.
    span: A spaCy Span, such as an entity in doc.ents
    target: A target phrase or iterable of phrases to check in span.lower_.
    regex (bool): Whether to search the span using a regular expression rather than
        a literal string. Default True.
    """
    if regex is True:
        if case_insensitive:
            func = lambda x: re.search(x, span.text, flags=re.IGNORECASE) is not None
        else:
            func = lambda x: re.search(x, span.text) is not None
    else:
        if case_insensitive:
            func = lambda x: x.lower() in span.lower_
        else:
            func = lambda x: x in span.text

    if isinstance(target, str):
        return func(target)

    # If it's an iterable, check if any of the strings are in sent
    for string in target:
        if func(string):
            return True
    return False


import re
import warnings

warnings.filterwarnings("once", "You are using a TargetRule with a regex pattern.*")


class RegexMatcher:
    def __init__(self, vocab, flags=re.IGNORECASE, resolve_start="left", resolve_end="right"):
        """Create a new RegexMatcher. The RegexMatcher is an alternative to spaCy's native
        Matcher and PhraseMatcher classes and allows matching based on typical regular expressions
        over the underlying doc text rather than spacy token attributes. This can be useful for
        allowing more traditional text matching methods, but can lead to issues if the matched spans
        in the text do not line up with spacy token boundaries. In this case, the RegexMatcher will
        by default resolve to the nearest token boundaries by expanding to the left and right.
        This behavior can be configured using `resolve_start` and `resolve_end`.
        To avoid this, consider using a list of dicts, such as in a spacy Matcher.
        For more information, see: https://spacy.io/usage/rule-based-matching

        Params:
            vocab (Vocab): spacy model vocabulary
            flags (int or None): Regular expression flags. Default re.IGNORECASE
            resolve_start (str): How to resolve if the start character index of a
                match does not align with spacy token boundaries.
                If 'left', will find the nearest token boundary to the left of the unmatched character index,
                    leading to a longer than expected span.
                If 'right', will find the nearest token boundary to the right of the unmatched character index,
                    leading to a shorter than expected span.
                Default 'left'.
            resolve_end (str):How to resolve if the end character index of a
                match does not align with spacy token boundaries.
                If 'left', will find the nearest token boundary to the left of the unmatched character index,
                    leading to a shorter than expected span.
                If 'right', will find the nearest token boundary to the right of the unmatched character index,
                    leading to a longer than expected span.
                Default 'right'.

        Returns:
            matches: A list of tuples containing:
                (match_id [int], start [int], end [int])

        Examples of resolve_start/resolve_end:
            In the string 'SERVICE: Radiology' the pattern 'ICE: Rad' would match in the middle of the tokens
            'SERVICE' and 'RADIOLOGY'. SpaCy would normally return None. The RegexMatcher will expand in the following ways:
            resolve_start='left': The resulting span will start at 'SERVICE' -> 'SERVICE: Radiology'
            resolve_start='right': The resulting span will start at ':' -> ': Radiology'
            resolve_end='left': The resulting span will end at ':': -> 'SERVICE:'
            resolve_end='right': The resulting span will end at 'RADIOLOGY' -> 'SERVICE: Radiology'

        """
        self.vocab = vocab
        self.flags = flags
        self.resolve_start = resolve_start
        self.resolve_end = resolve_end
        self._patterns = {}
        self._callbacks = {}
        self.labels = set()
        self._rule_item_mapping = dict()

    def add(self, match_id, patterns, on_match=None):
        warnings.warn(
            "You are using a TargetRule with a regex pattern, which is not "
            "natively supported in spacy and may lead to unexpected match spans. "
            "Consider using a list of dicts pattern instead. "
            "See https://spacy.io/usage/rule-based-matching",
            RuntimeWarning,
        )
        if match_id not in self.vocab:
            self.vocab.strings.add(match_id)
        self._patterns.setdefault(self.vocab.strings[match_id], [])
        for pattern in patterns:
            self._patterns[self.vocab.strings[match_id]].append(re.compile(pattern, flags=self.flags))
            self._callbacks[self.vocab.strings[match_id]] = on_match

    def get(self, key):
        return self._patterns.get(self.vocab.strings[key], [])

    def __call__(self, doc):
        matches = []
        for (match_id, patterns) in self._patterns.items():
            for pattern in patterns:
                on_match = self._callbacks[match_id]
                for re_match in pattern.finditer(doc.text_with_ws):
                    span = doc.char_span(re_match.start(), re_match.end())
                    if span is None:
                        start = get_token_for_char(doc, re_match.start(), resolve=self.resolve_start)
                        end = get_token_for_char(doc, re_match.end(), resolve=self.resolve_end)
                        if end is None:
                            end_index = len(doc)
                        else:
                            end_index = end.i
                        span = doc[start.i : end_index]
                    # If it's an empty span, then that means that the token resolution
                    # must have resulted in no tokens being included.
                    # Don't add the match
                    if len(span):
                        match = (match_id, span.start, span.end)
                        matches.append(match)
                    # If a callback function was defined,
                    # call it according to the spaCy API:
                    # https://spacy.io/usage/rule-based-matching#on_match
                    if on_match is not None:
                        on_match(self, doc, len(matches) - 1, matches)

        return matches


def get_token_for_char(doc, char_idx, resolve="left"):
    if char_idx < 0:
        return ValueError("char_idx must be > 0")
    if char_idx > len(doc.text_with_ws) - 1:
        return ValueError("char_idx {0} is out of range for text with length {1}".format(char_idx, len(doc.text_with_ws)))
    for i, token in enumerate(doc):
        if char_idx > token.idx:
            continue
        if char_idx == token.idx:
            return token
        if char_idx < token.idx:
            if resolve == "left":
                return doc[i - 1]
            elif resolve == "right":
                return doc[i]
            else:
                raise ValueError("resolve must be either 'left' or 'right'")
    # Otherwise, we've reached the end of the doc, so this must be the final token
    # If resolving to the left, return the final token
    # If resolving to the right, return None, meaning it should go to the end of the doc
    if resolve == "left":
        return doc[-1]
    if resolve == "right":
        return None



# from nlp_tools.utils import prune_overlapping_matches

from spacy.tokens import Token
from spacy.tokens import Span
from .target_rule import TargetRule
from ..common.medspacy_matcher import MedspacyMatcher



class TargetMatcher:
    """TargetMatcher is a component for advanced direction-based text extraction.
    Rules are defined using medspacy.target_matcher.TargetRule.
    """

    name = "target_matcher"

    def __init__(self, nlp, add_ents=True, phrase_matcher_attr="LOWER"):
        """Create a new TargetMatcher.
        Params:
            nlp: A spaCy Language model.
            add_ents: Whether to add extracted spans to doc.ents. If True, will attempt to add
                resulting entities to existing ents. If resulting spans conflict with existing ents,
                the existing entities will take precedence.
                If False, will return a list of spans.
            phrase_matcher_attr: The token attribute to use for PhraseMatcher for rules where `pattern` is None.
                Default is "LOWER".
        """
        self.nlp = nlp
        self.add_ents = add_ents
        self._rules = list()

        self.labels = set()

        self.matcher = MedspacyMatcher(nlp, phrase_matcher_attr=phrase_matcher_attr)
        self._rule_item_mapping = self.matcher._rule_item_mapping

    def add(self, rules):
        """Add a list of targetRules to the matcher."""
        if not isinstance(rules, list):
            rules = [rules]
        self._rules += rules
        self.matcher.add(rules)
        for rule in rules:
            if not isinstance(rule, TargetRule):
                raise ValueError("Rules must be TargetRule, not", type(rule))

    @property
    def rules(self):
        return self._rules

    def __call__(self, doc):
        """Call TargetMatcher on a doc. If `add_ents=True`, then matched
        spans will be merged in to doc.ents and `doc` will be returned.
        If `add_ents=False`, then matched spans will be returned as a list,
        in which case this cannot be used as part of a spaCy pipeline, which
        requires each component to return the doc, but can be used as a standalone matcher.

        In addition to extracting spans of text and setting labels, TargetRules
        can also define setting custom attributes and metadata. Additionally,
        each resulting span has an attribute span._.target_rule which maps
        a span to the TargetRule which set it.
        """
        matches = self.matcher(doc)
        spans = []
        for (rule_id, start, end) in matches:
            rule = self._rule_item_mapping[self.nlp.vocab.strings[rule_id]]
            span = Span(doc, start=start, end=end, label=rule.category)
            span._.target_rule = rule
            if rule.attributes is not None:
                for (attribute, value) in rule.attributes.items():
                    try:
                        setattr(span._, attribute, value)
                    except AttributeError as e:
                        raise e
            spans.append(span)
        if self.add_ents is True:
            for span in spans:
                try:
                    doc.ents += (span,)
                # spaCy will raise a value error if the token in span are already
                # part of an entity (ie., as part of an upstream component)
                # In that case, let the existing span supersede this one
                except ValueError as e:
                    # raise e
                    pass
            return doc
        else:
            return spans



from ..common.base_rule import BaseRule


class SectionRule(BaseRule):

    _ALLOWED_KEYS = {"literal", "pattern", "category", "metadata", "parents", "parent_required"}

    def __init__(
        self, literal, category, pattern=None, on_match=None, metadata=None, parents=[], parent_required=False,
    ):
        """Class for defining rules for extracting entities from text using TargetMatcher.
        Params:
            literal (str): The actual string of a concept. If pattern is None,
                this string will be lower-cased and matched to the lower-case string.
                If `pattern` is not None, this argument will not be used for actual matching
                but can be used as a reference as the direction name.
            category (str): The semantic class of the matched span. This corresponds to the `label_`
                attribute of an entity.
            pattern (list, str, or None): A pattern to use for matching rather than `literal`.
                If a list, will use spaCy dictionary pattern matching to match using token attributes.
                See https://spacy.io/usage/rule-based-matching.
                If a string, will use regular expression matching on the underlying text of a doc.
                Note that regular-expression matching is not natively supported by spaCy and could
                result in unexpected matched spans if match boundaries do not align with token boundaries.
                If None, `literal` will be matched exactly.
            on_match (callable or None): An optional callback function or other callable which takes 4 arguments:
                (matcher, doc, i, matches)
                For more information, see https://spacy.io/usage/rule-based-matching#on_match
            meta (dict or None): Optional dictionary of metadata.
            parents (list or None): a list of candidate parents for determining subsections
            parent_required (bool): whether a parent is required for the section to exist in the final output
        """
        super().__init__(literal, category, pattern, on_match, metadata)
        self.parents = parents
        if parent_required:
            if not parents:
                raise ValueError(
                    "Jsonl file incorrectly formatted for pattern name {0}. If parents are required, then at least one parent must be specified.".format(
                        category
                    )
                )
        self.parent_required = parent_required

    @classmethod
    def from_json(cls, filepath):
        """Read in a lexicon of modifiers from a JSON file.

        Args:
            filepath: the .json file containing modifier rules

        Returns:
            section_rules: a list of SectionRule objects
        Raises:
            KeyError: if the dictionary contains any keys other than
                those accepted by SectionRule.__init__
        """
        import json

        with open(filepath) as file:
            target_data = json.load(file)
        section_rules = []
        for data in target_data["section_rules"]:
            section_rules.append(SectionRule.from_dict(data))
        return section_rules

    @classmethod
    def from_dict(cls, rule_dict):
        """Reads a dictionary into a SectionRule list. Used when reading from a json file.

        Args:
            item_dict: the dictionary to convert

        Returns:
            item: the SectionRule created from the dictionary

        Raises:
            ValueError: if the json is invalid
        """
        keys = set(rule_dict.keys())
        invalid_keys = keys.difference(cls._ALLOWED_KEYS)
        if invalid_keys:
            msg = "JSON object contains invalid keys: {0}.\n" "Must be one of: {1}".format(invalid_keys, cls._ALLOWED_KEYS)
            raise ValueError(msg)
        rule = SectionRule(**rule_dict)
        return rule

    @classmethod
    def to_json(cls, target_rules, filepath):
        """Writes SectionRules to a json file.

        Args:
            target_rules: a list of TargetRules that will be written to a file.
            filepath: the .json file to contain modifier rules
        """
        import json

        data = {"target_rules": [rule.to_dict() for rule in target_rules]}
        with open(filepath, "w") as file:
            json.dump(data, file, indent=4)

    def to_dict(self):
        """Converts TargetRules to a python dictionary. Used when writing target rules to a json file.

        Returns:
            rule_dict: the dictionary containing the TargetRule info.
        """
        rule_dict = {}
        for key in self._ALLOWED_KEYS:
            rule_dict[key] = self.__dict__.get(key)
        return rule_dict

    def __repr__(self):
        return f"""SectionRule(literal="{self.literal}", category="{self.category}", pattern={self.pattern}, on_match={self.on_match}, parents={self.parents}, parent_required={self.parent_required})"""



class BaseRule:
    def __init__(self, literal, category, pattern=None, on_match=None, metadata=None):
        """Base class for medspacy rules such as TargetRule and ConTextRule."""
        self.literal = literal
        self.category = category
        self.pattern = pattern
        self.on_match = on_match
        self.metadata = metadata



"""This module will set extension attributes and methods for medspaCy. Examples include custom methods like span._.window()"""
from spacy.tokens import Doc, Span, Token
from .common .util import span_contains

def set_extensions():
    "Set custom medspaCy extensions for Token, Span, and Doc classes."
    set_token_extensions()
    set_span_extensions()
    set_doc_extensions()

def set_token_extensions():
    for attr, attr_info in _token_extensions.items():
        try:
            Token.set_extension(attr, **attr_info)
        except ValueError as e: # If the attribute has already set, this will raise an error
            pass

def set_span_extensions():
    for attr, attr_info in _span_extensions.items():
        try:
            Span.set_extension(attr, **attr_info)
        except ValueError as e: # If the attribute has already set, this will raise an error
            pass

def set_doc_extensions():
    for attr, attr_info in _doc_extensions.items():
        try:
            Doc.set_extension(attr, **attr_info)
        except ValueError as e: # If the attribute has already set, this will raise an error
            pass

def get_extensions():
    "Get a list of medspaCy extensions for Token, Span, and Doc classes."
    return {
        "Token": get_token_extensions(),
        "Span": get_span_extensions(),
        "Doc": get_doc_extensions()
    }

def get_token_extensions():
    return _token_extensions

def get_span_extensions():
    return _span_extensions

def get_doc_extensions():
    return _doc_extensions

def get_window_token(token, n=1, left=True, right=True):
    """Get a Span of a window of text containing a token.
    Args:
        n (int): Number of tokens on each side of token to return.
            Default 1.
        left (bool): Whether to include the span preceding token.
            Default True.
        right (bool): Whether to include the span following token.
            Default True.
    Returns:
        a spaCy Span
    """
    if left:
        start = max((token.i-n, 0))
    else:
        start = token.i
    if right:
        end = min((token.i+n+1, len(token.doc)))
    else:
        end = token.i+1
    return token.doc[start:end]

def get_window_span(span, n=1, left=True, right=True):
    """Get a Span of a window of text containing a span.
    Args:
        n (int): Number of tokens on each side of a span to return.
            Default 1.
        left (bool): Whether to include the span precedinga span.
            Default True.
        right (bool): Whether to include the span following a span.
            Default True.
    Returns:
        a spaCy Span
    """
    if left:
        start = max((span.start-n, 0))
    else:
        start = span.start
    if right:
        end = min((span.end+n, len(span.doc)))
    else:
        end = span.end
    return span.doc[start:end]

def get_context_attributes(span):
    """Return a dict of all ConText assertion attributes (is_negated, is_historical, etc.)
    and their values.
    """
    attr_dict = dict()
    for attr in _context_attributes:
        attr_dict[attr] = span._.get(attr)
    return attr_dict

def any_context_attribute(span):
    "Return True if any of the ConText assertion attributes (is_negated, is_historical, etc.) is True."
    return any(span._.context_attributes.values())

def get_section_title_spans(doc):
    return [section.title_span for section in doc._.sections]

def get_section_categories(doc):
    return [section.category for section in doc._.sections]

def get_section_parents(doc):
    return [section.parent for section in doc._.sections]

def get_section_spans(doc):
    return [section.section_span for section in doc._.sections]

def get_section_body_spans(doc):
    return [section.body_span for section in doc._.sections]

def get_section_span_token(token):
    if token._.section is None:
        return None
    return token._.section.section_span

def get_section_category_token(token):
    if token._.section is None:
        return None
    return token._.section.category

def get_section_title_span_token(token):
    if token._.section is None:
        return None
    return token._.section.title_span

def get_section_body_span_token(token):
    if token._.section is None:
        return None
    return token._.section.body_span

def get_section_parent_token(token):
    if token._.section is None:
        return None
    return token._.section.parent

def get_section_rule_token(token):
    if token._.section is None:
        return None
    return token._.section.rule


_token_extensions = {
   "window": {"method": get_window_token},
    "section": {"default": None},
   "section_span": {"getter": get_section_span_token},
   "section_category": {"getter": get_section_category_token},
   "section_title": {"getter": get_section_title_span_token},
   "section_body": {"getter": get_section_body_span_token},
   "section_parent": {"getter": get_section_parent_token},
   "section_rule": {"getter": get_section_rule_token},
}

_context_attributes = {
    "is_negated": {"default": False},
    "is_historical": {"default": False},
    "is_hypothetical": {"default": False},
    "is_family": {"default": False},
    "is_uncertain": {"default": False},
}


_span_extensions = {
    "window": {"method": get_window_span},
    "context_attributes": {"getter": get_context_attributes},
    "any_context_attributes": {"getter": any_context_attribute},
    "section": {"getter":lambda x: x[0]._.section},
    "section_span": {"getter":lambda x: x[0]._.section_span},
    "section_category": {"getter":lambda x: x[0]._.section_category},
    "section_title": {"getter":lambda x: x[0]._.section_title},
    "section_body": {"getter":lambda x: x[0]._.section_body},
    "section_parent": {"getter":lambda x: x[0]._.section_parent},
    "section_rule": {"getter":lambda x: x[0]._.section_rule},
    "contains": {"method": span_contains},
    "target_rule": {"default": None},
    **_context_attributes

}

_doc_extensions = {
    "sections": {"default": list()},
    "section_titles": {"getter": get_section_title_spans},
    "section_categories": {"getter": get_section_categories},
    "section_spans": {"getter": get_section_spans},
    "section_parents": {"getter": get_section_parents},
    "section_bodies": {"getter": get_section_body_spans}
}






from spacy.matcher import Matcher, PhraseMatcher
from .regex_matcher import RegexMatcher
from .base_rule import BaseRule

from spacy.tokens import Span


class MedspacyMatcher:
    """MedspacyMatcher is a class which combines spaCy's Matcher and PhraseMatcher classes
    along with medspaCy's RegexMatcher and acts as one single matcher using 3 different types of rules:
        - Exact phrases
        - List of dictionaries for matching on token attributes (see https://spacy.io/usage/rule-based-matching#matcher)
        - Regular expression matches. Note that regular-expression matching is not natively supported by spaCy and could
                result in unexpected matched spans if match boundaries do not align with token boundaries.
    Rules can be defined by any class which inherits from medspacy.common.BaseRule, such as:
        medspacy.target_matcher.TargetRule
        medspacy.context.ConTextRule
    """

    name = "medspacy_matcher"

    def __init__(self, nlp, phrase_matcher_attr="LOWER"):
        """Create a MedspacyMatcher.
        Params:
            nlp: A spaCy Language model.
            phrase_matcher_attr: The attribute to use for spaCy's PhraseMatcher (default is 'LOWER')
        """
        self.nlp = nlp
        self._rule_ids = set()
        self._rules = list()
        self.labels = set()
        self._rule_item_mapping = dict()

        self.matcher = Matcher(self.nlp.vocab)
        self.phrase_matcher = PhraseMatcher(self.nlp.vocab, attr=phrase_matcher_attr)
        self.regex_matcher = RegexMatcher(self.nlp.vocab)

    def add(self, rules):
        """Add a list of rules to the matcher. Rules must inherit from medspacy.common.BaseRule,
        such as medspacy.target_matcher.TargetRule or medspacy.context.ConTextRule."""
        i = len(self._rules)
        self._rules += rules

        for rule in rules:
            if not isinstance(rule, BaseRule):
                raise ValueError(
                    "Rules must inherit from medspacy.common.BaseRule, " "such as medspacy.target_matcher.TargetRule."
                )
            self.labels.add(rule.category)
            rule_id = f"{rule.category}_{i}"
            rule._rule_id = rule_id
            self._rule_item_mapping[rule_id] = rule
            if rule.pattern is not None:
                # If it's a string, add a RegEx
                if isinstance(rule.pattern, str):
                    self.regex_matcher.add(rule_id, [rule.pattern], rule.on_match)
                # If it's a list, add a pattern dictionary
                elif isinstance(rule.pattern, list):
                    self.matcher.add(rule_id, [rule.pattern], on_match=rule.on_match)
                else:
                    raise ValueError(
                        "The pattern argument must be either a string or a list, not {0}".format(type(rule.pattern))
                    )
            else:
                self.phrase_matcher.add(rule_id, [self.nlp.make_doc(rule.literal.lower())], on_match=rule.on_match)
            i += 1

    def __call__(self, doc):
        """Call MedspacyMatcher on a doc and return a single list of matches. If rules result in multiple
        overlapping matches, the longest will be returned."""
        matches = self.matcher(doc)
        matches += self.phrase_matcher(doc)
        matches += self.regex_matcher(doc)
        matches = prune_overlapping_matches(matches)
        return matches


def prune_overlapping_matches(matches, strategy="longest"):
    if strategy != "longest":
        raise NotImplementedError()

    # Make a copy and sort
    unpruned = sorted(matches, key=lambda x: (x[1], x[2]))
    pruned = []
    num_matches = len(matches)
    if num_matches == 0:
        return matches
    curr_match = unpruned.pop(0)

    while True:
        if len(unpruned) == 0:
            pruned.append(curr_match)
            break
        next_match = unpruned.pop(0)

        # Check if they overlap
        if overlaps(curr_match, next_match):
            # Choose the larger span
            longer_span = max(curr_match, next_match, key=lambda x: (x[2] - x[1]))
            pruned.append(longer_span)
            if len(unpruned) == 0:
                break
            curr_match = unpruned.pop(0)
        else:
            pruned.append(curr_match)
            curr_match = next_match
    # Recursive base point
    if len(pruned) == num_matches:
        return pruned
    # Recursive function call
    else:
        return prune_overlapping_matches(pruned)


def overlaps(a, b):
    if _match_overlaps(a, b) or _match_overlaps(b, a):
        return True
    return False


def _match_overlaps(a, b):
    _, a_start, a_end = a
    _, b_start, b_end = b
    if a_start >= b_start and a_start < b_end:
        return True
    if a_end > b_start and a_end <= b_end:
        return True
    return False


def matches_to_spans(doc, matches, set_label=True):
    spans = []
    for (rule_id, start, end) in matches:
        if set_label:
            label = doc.vocab.strings[rule_id]
        else:
            label = None
        spans.append(Span(doc, start=start, end=end, label=label))
    return spans



from setuptools import setup, find_packages
from sys import platform

# read the contents of the README file
import os
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

additional_installs = []
if platform.startswith('win'):
    print('Not installing QuickUMLS for Windows since it currently requires conda (as opposed to just pip)')
else:
    # Using a trick from StackOverflow to set an impossibly high version number
    # to force getting latest from GitHub as opposed to PyPi
    # since QuickUMLS has not made a release with some recent MedSpacy contributions...
    quickumls_package = 'medspacy_quickumls>=2.4.1'
    additional_installs.append(quickumls_package)
    print('Attempting to install quickumls package: {}'.format(quickumls_package))

# function to recursively get files for resourcee
def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


# get all files recursively from /resources
resource_files = package_files('./resources')

def get_version():
    """Load the version from version.py, without importing it.
    This function assumes that the last line in the file contains a variable defining the
    version string with single quotes.
    """
    try:
        with open('medspacy/_version.py', 'r') as f:
            return f.read().split('\n')[0].split('=')[-1].replace('\'', '').strip()
    except IOError:
        raise IOError

setup(
    name="medspacy",
    version=get_version(),
    description="Library for clinical NLP with spaCy.",
    author="medSpaCy",
    author_email="medspacy.dev@gmail.com",
    packages=find_packages(),
    install_requires=[
        # NOTE: spacy imports numpy to bootstrap its own setup.py in 2.3.2
        "spacy>=2.3.0,<=2.3.2",
        "PyRuSH>=1.0.3.5",
        "jsonschema"
    ] + additional_installs,
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_data={"medspacy": resource_files},
)



from .sectionizer import Sectionizer, Section
# from .text_sectionizer import TextSectionizer
from .section_rule import SectionRule
from .section import Section
from .util import section_patterns_to_rules
from ._version import __version__

__all__ = ["Sectionizer",
           # "TextSectionizer",
           "SectionRule",
           "Section"
           ]



NEWLINE_PATTERN = r"[\n\r]+[\s]*$"

def is_start_line(idx, doc, pattern):
    # If it's the start of the doc, return True
    if idx == 0:
        return True
    # Otherwise, check if the preceding token ends with newlines
    preceding_text = doc[idx - 1].text_with_ws
    return pattern.search(preceding_text) is not None


def is_end_line(idx, doc, pattern):
    # If it's the end of the doc, return True
    if idx == len(doc) - 1:
        return True

    # Check if either the token has trailing newlines,
    # or if the next token is a newline
    text = doc[idx].text_with_ws
    if pattern.search(text) is not None:
        return True
    following_text = doc[idx + 1].text_with_ws
    return pattern.search(following_text) is not None

def section_patterns_to_rules(patterns):
    """Convert a list of dictionary-based patterns of the old Sectionizer API
    to a list of SectionRules.
    """
    _ALLOWED_KEYS = {"category", "metadata", "parents", "parent_required"}
    from .section_rule import SectionRule

    rules = []
    for pattern in patterns:
        refactored = {}
        if isinstance(pattern["pattern"], str):
            refactored["literal"] = pattern["pattern"]
            refactored["pattern"] = None
        elif isinstance(pattern["pattern"], list):
            refactored["literal"] = str(pattern["pattern"])
            refactored["pattern"] = pattern["pattern"]

        refactored["category"] = pattern["section_title"]
        for key in _ALLOWED_KEYS:
            if key in pattern:
                refactored[key] = pattern[key]
        rule = SectionRule(**refactored)
        rules.append(rule)
    return rules
